import typing as t
from django.test import TestCase
from rest_framework.test import APIClient
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from .models import (
    get_tenant_model,
    Tenant,
    Member,
    Permission,
)


class SaasTestCase(TestCase):
    fixtures = [
        'test_data.yaml',
    ]

    client: APIClient
    client_class = APIClient

    tenant_id: int = 1
    user_id: int = 0

    ADMIN_USER_ID = 1
    STAFF_USER_ID = 2
    OWNER_USER_ID = 3
    GUEST_USER_ID = 4
    EMPTY_USER_ID = 5

    def setUp(self) -> None:
        if self.tenant_id:
            self.client.credentials(
                HTTP_X_TENANT_ID=str(self.tenant_id),
            )

    @property
    def tenant(self):
        return self.get_tenant()

    @property
    def user(self):
        return self.get_user()

    def get_user(self, pk: t.Optional[int] = None) -> User:
        if pk is None:
            pk = self.user_id
        return get_user_model().objects.get(pk=pk)

    def get_tenant(self, pk: t.Optional[int] = None) -> Tenant:
        if pk is None:
            pk = self.tenant_id
        return get_tenant_model().objects.get(pk=pk)

    def get_user_member(self):
        user = self.get_user()
        tenant = self.get_tenant()
        return Member.objects.get(user=user, tenant=tenant)

    def add_user_perms(self, *perms: str):
        member = self.get_user_member()
        for obj in Permission.objects.filter(name__in=perms).all():
            member.permissions.add(obj)
        return member

    def force_login(self, user_id: int = None):
        if user_id is None:
            user = self.get_user()
        else:
            user = self.get_user(user_id)
        self.client.force_login(user)



from rest_framework.mixins import (
    RetrieveModelMixin,
    CreateModelMixin,
    UpdateModelMixin,
    ListModelMixin,
)
from rest_framework.request import Request
from rest_framework.permissions import IsAuthenticated
from rest_framework.settings import api_settings
from ..drf.views import AuthenticatedEndpoint, TenantEndpoint
from ..models import get_tenant_model
from ..serializers.tenant import TenantSerializer, TenantUpdateSerializer

__all__ = [
    'SelectedTenantEndpoint',
    'TenantListEndpoint',
    'TenantItemEndpoint',
]


class SelectedTenantEndpoint(RetrieveModelMixin, TenantEndpoint):
    serializer_class = TenantSerializer
    resource_name = 'tenant'
    tenant_id_field = 'pk'

    def get_queryset(self):
        return get_tenant_model().objects.all()

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        obj = self.get_object_or_404(queryset)
        self.check_object_permissions(self.request, obj)
        return obj

    def get(self, request: Request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)


class TenantListEndpoint(CreateModelMixin, ListModelMixin, AuthenticatedEndpoint):
    serializer_class = TenantSerializer
    pagination_class = None
    permission_classes = [IsAuthenticated] + api_settings.DEFAULT_PERMISSION_CLASSES

    def get_queryset(self):
        return get_tenant_model().objects.filter(owner=self.request.user).all()

    def get(self, request: Request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request: Request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)


class TenantItemEndpoint(RetrieveModelMixin, UpdateModelMixin, TenantEndpoint):
    serializer_class = TenantUpdateSerializer
    queryset = get_tenant_model().objects.all()
    filter_backends = []
    resource_name = 'tenant'
    tenant_id_field = 'pk'

    def get(self, request: Request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def patch(self, request: Request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)



from rest_framework import serializers
from .permission import PermissionSerializer
from ..drf.serializers import RelatedSerializerField
from ..models import Role


class RoleSerializer(serializers.ModelSerializer):
    permissions = RelatedSerializerField(PermissionSerializer, many=True)

    class Meta:
        model = Role
        fields = ['name', 'description', 'permissions']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import os
import sys


def main():
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'demo.settings')
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            'available on your PYTHONPATH environment variable? Did you '
            'forget to activate a virtual environment?'
        ) from exc
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



from django.core.cache import caches, BaseCache
from django.utils.connection import ConnectionProxy
from ..settings import saas_settings

cache: BaseCache = ConnectionProxy(caches, saas_settings.DB_CACHE_ALIAS)  # type: ignore[assignment]


__all__ = ['cache']



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend as BaseModelBackend
from saas_base.models import UserEmail

__all__ = ['ModelBackend']

UserModel = get_user_model()


class ModelBackend(BaseModelBackend):
    @staticmethod
    def _get_user_by_username(username: str):
        try:
            return UserModel.objects.get(username=username)
        except UserModel.DoesNotExist:
            if '@' not in username:
                return None
            try:
                obj = UserEmail.objects.select_related('user').get(email=username)
                if obj.verified and obj.primary:
                    return obj.user
            except UserEmail.DoesNotExist:
                return None
        return None

    def authenticate(self, request, username=None, password=None, **kwargs):
        if username is None or password is None:
            return

        user = self._get_user_by_username(username)
        if user is None:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a nonexistent user (#20760).
            UserModel().set_password(password)
        elif user.check_password(password) and self.user_can_authenticate(user):
            return user



import uuid
from django.utils import timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Permission',
            fields=[
                ('name', models.CharField(editable=False, max_length=100, primary_key=True, serialize=False)),
                ('description', models.CharField(blank=True, max_length=480)),
                ('internal', models.BooleanField(db_index=True, default=False)),
                ('created_at', models.DateTimeField(default=timezone.now)),
            ],
            options={
                'verbose_name': 'permission',
                'verbose_name_plural': 'permissions',
                'db_table': 'saas_permission',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Tenant',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=140)),
                ('slug', models.SlugField(help_text='Identity of the tenant, e.g. <slug>.example.com', unique=True)),
                ('region', models.CharField(blank=True, default='', max_length=24)),
                ('environment', models.CharField(blank=True, default='', max_length=48)),
                ('expires_at', models.DateTimeField(blank=True, null=True)),
                ('created_at', models.DateTimeField(db_index=True, default=timezone.now)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('owner', models.ForeignKey(null=True, on_delete=models.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'tenant',
                'verbose_name_plural': 'tenants',
                'db_table': 'saas_tenant',
                'ordering': ['created_at'],
                'abstract': False,
                'swappable': 'SAAS_TENANT_MODEL',
            },
        ),
        migrations.CreateModel(
            name='Group',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(db_index=True, max_length=100)),
                ('managed', models.BooleanField(default=False)),
                ('created_at', models.DateTimeField(db_index=True, default=timezone.now)),
                ('tenant', models.ForeignKey(on_delete=models.CASCADE, to=settings.SAAS_TENANT_MODEL)),
                ('permissions', models.ManyToManyField(blank=True, to='saas_base.permission')),
            ],
            options={
                'verbose_name': 'group',
                'verbose_name_plural': 'groups',
                'db_table': 'saas_group',
                'ordering': ['created_at'],
                'unique_together': {('tenant', 'name')},
            },
        ),
        migrations.CreateModel(
            name='UserEmail',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('email', models.EmailField(max_length=254, unique=True)),
                ('verified', models.BooleanField(default=False)),
                ('primary', models.BooleanField(db_index=True, default=False)),
                ('created_at', models.DateTimeField(db_index=True, default=timezone.now)),
                (
                    'user',
                    models.ForeignKey(on_delete=models.CASCADE, related_name='emails', to=settings.AUTH_USER_MODEL),
                ),
            ],
            options={
                'verbose_name': 'email',
                'verbose_name_plural': 'emails',
                'db_table': 'saas_auth_email',
                'ordering': ['created_at'],
            },
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('invite_email', models.EmailField(blank=True, max_length=254, null=True)),
                (
                    'status',
                    models.SmallIntegerField(choices=[(0, 'request'), (1, 'waiting'), (2, 'active')], default=0),
                ),
                ('created_at', models.DateTimeField(db_index=True, default=timezone.now)),
                (
                    'groups',
                    models.ManyToManyField(
                        blank=True,
                        help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.',
                        to='saas_base.group',
                    ),
                ),
                (
                    'inviter',
                    models.ForeignKey(
                        blank=True, null=True, on_delete=models.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL
                    ),
                ),
                ('tenant', models.ForeignKey(on_delete=models.CASCADE, to=settings.SAAS_TENANT_MODEL)),
                (
                    'user',
                    models.ForeignKey(blank=True, null=True, on_delete=models.SET_NULL, to=settings.AUTH_USER_MODEL),
                ),
                (
                    'permissions',
                    models.ManyToManyField(
                        blank=True, help_text='Specific permissions for this user.', to='saas_base.permission'
                    ),
                ),
            ],
            options={
                'verbose_name': 'member',
                'verbose_name_plural': 'members',
                'db_table': 'saas_member',
                'ordering': ['-created_at'],
                'unique_together': {('tenant', 'user')},
            },
        ),
    ]



from django.urls import path
from ..endpoints.user_members import (
    UserMemberListEndpoint,
    UserMemberItemEndpoint,
)

urlpatterns = [
    path('', UserMemberListEndpoint.as_view()),
    path('<pk>/', UserMemberItemEndpoint.as_view()),
]



from django.urls import path
from ..endpoints.user_emails import (
    UserEmailListEndpoint,
    UserEmailItemEndpoint,
    AddUserEmailRequestEndpoint,
    AddUserEmailConfirmEndpoint,
)

urlpatterns = [
    path('', UserEmailListEndpoint.as_view()),
    path('add/request/', AddUserEmailRequestEndpoint.as_view()),
    path('add/confirm/', AddUserEmailConfirmEndpoint.as_view()),
    path('<pk>/', UserEmailItemEndpoint.as_view()),
]



from django.utils import timezone
from django.utils.translation import gettext as _
from rest_framework.permissions import BasePermission, SAFE_METHODS
from rest_framework.request import Request
from rest_framework.exceptions import PermissionDenied
from saas_base.models import get_tenant_model, get_cached_tenant, Member
from saas_base.settings import saas_settings

__all__ = [
    'IsTenantOwner',
    'IsTenantOwnerOrReadOnly',
    'IsTenantActive',
    'IsTenantActiveOrReadOnly',
    'HasResourcePermission',
    'HasResourcePermissionOrReadOnly',
    'HasResourceScope',
]

TenantModel = get_tenant_model()

http_method_actions = {
    'GET': 'read',
    'HEAD': 'read',
    'POST': 'write',
    'PUT': 'write',
    'PATCH': 'write',
    'DELETE': 'admin',
}


class BaseTenantPermission(BasePermission):
    tenant_id_field = 'tenant_id'

    @staticmethod
    def check_tenant_permission(request, view, tenant_id):
        return True

    def has_permission(self, request: Request, view):
        tenant_id = getattr(request, 'tenant_id', None)
        return self.check_tenant_permission(request, view, tenant_id)

    def has_object_permission(self, request, view, obj):
        tenant_id_field = getattr(view, 'tenant_id_field', self.tenant_id_field)
        if tenant_id_field is None:
            return True
        tenant_id = getattr(obj, tenant_id_field, None)
        return self.check_tenant_permission(request, view, tenant_id)


class IsTenantOwner(BaseTenantPermission):
    """The authenticated user is the tenant owner."""

    @staticmethod
    def check_tenant_permission(request, view, tenant_id):
        if not tenant_id:
            return False

        tenant = get_cached_tenant(tenant_id, request._request)
        if not tenant:
            return False

        return request.user.pk == tenant.owner_id


class IsTenantOwnerOrReadOnly(IsTenantOwner):
    """The authenticated user is the tenant owner, or is a read-only request."""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if request.method in SAFE_METHODS:
            return True
        return super().has_object_permission(request, view, obj)


class IsTenantActive(BaseTenantPermission):
    """The requested tenant is not expired."""

    @staticmethod
    def check_tenant_permission(request, view, tenant_id):
        tenant = get_cached_tenant(tenant_id, request._request)
        if not tenant:
            return False

        if not tenant.expires_at:
            return True
        # tenant not expired
        if tenant.expires_at < timezone.now():
            raise PermissionDenied(_('This tenant is expired.'))
        return True


class IsTenantActiveOrReadOnly(IsTenantActive):
    """The requested tenant is not expired, or is a read-only request."""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if request.method in SAFE_METHODS:
            return True
        return super().has_object_permission(request, view, obj)


class HasResourcePermission(BaseTenantPermission):
    """The authenticated user is a member of the tenant, and the user
    has the given resource permission.
    """

    @staticmethod
    def get_resource_permissions(view, method):
        handler = getattr(view, method.lower(), None)
        if not handler:
            return None

        permissions = getattr(handler, '_resource_permissions', None)
        if permissions:
            return permissions

        resource = getattr(view, 'resource_name', None)
        if not resource:
            return None

        action = getattr(view, 'resource_action', None)
        if not action:
            method_actions = getattr(view, 'resource_http_method_actions', http_method_actions)
            action = method_actions.get(method)

        permission = saas_settings.PERMISSION_NAME_FORMATTER.format(
            resource=resource,
            action=action,
        )
        return [permission]

    @classmethod
    def check_tenant_permission(cls, request: Request, view, tenant_id):
        resource_permissions = cls.get_resource_permissions(view, request.method)
        if not resource_permissions:
            return True

        if not tenant_id:
            return False

        tenant = get_cached_tenant(tenant_id, request._request)
        if not tenant:
            return False

        # Tenant owner has all permissions
        if tenant.owner_id == request.user.pk:
            return True

        try:
            member = Member.objects.get_by_natural_key(tenant.pk, request.user.pk)
            if not member.is_active:
                return False
        except Member.DoesNotExist:
            return False

        perms = member.get_all_permissions()
        if not perms:
            return False

        return bool(set(resource_permissions) & perms)


class HasResourcePermissionOrReadOnly(HasResourcePermission):
    """The authenticated user has the tenant permission, or is a read-only request."""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if request.method in SAFE_METHODS:
            return True
        return super().has_object_permission(request, view, obj)


class HasResourceScope(BasePermission):
    """The request token contains the given resource scopes."""

    @staticmethod
    def get_resource_scopes(view, method):
        if hasattr(view, 'get_resource_scopes'):
            resource_scopes = view.get_resource_scopes(method)
        elif hasattr(view, 'resource_scopes'):
            resource_scopes = view.resource_scopes
        else:
            resource_scopes = None
        return resource_scopes

    def has_permission(self, request: Request, view):
        resource_scopes = self.get_resource_scopes(view, request.method)
        if not resource_scopes:
            return True

        # not using token for authentication
        if request.auth is None:
            return True

        # not using token with scope for authentication
        if not hasattr(request.auth, 'scope'):
            return True

        scope = getattr(request.auth, 'scope', '')
        # this token accepts all scopes
        if scope == '__all__':
            return True

        token_scopes = set(scope.split())
        for rs in resource_scopes:
            if set(rs.split()).issubset(token_scopes):
                return True
        return False



import requests
from rest_framework.request import Request
from saas_base.functions.ip import get_client_ip
from .base import Rule


class Turnstile(Rule):
    API_URL = 'https://challenges.cloudflare.com/turnstile/v0/siteverify'
    DEFAULT_RESPONSE_FIELD = 'cf-turnstile-response'

    def bad_request(self, request: Request):
        token = self.get_response_field_value(request)
        if not token:
            return True

        secret = self.options.get('secret')
        ip = get_client_ip(request)
        data = {'secret': secret, 'remoteip': ip, 'response': token}
        resp = requests.post(self.API_URL, data=data, timeout=5)
        resp.raise_for_status()
        data = resp.json()
        return not data.get('success')



import typing as t
import css_inline
from django.template import loader
from django.tasks import task, TaskResult
from django.core.mail import EmailMultiAlternatives
from saas_base.settings import saas_settings
from saas_base.signals import mail_queued


def send_template_email(
    template_id: str,
    subject: str,
    recipients: t.List[str],
    context: t.Dict[str, t.Any],
    from_email: t.Optional[str] = None,
    headers: t.Optional[t.Dict[str, str]] = None,
    reply_to: t.Optional[str] = None,
):
    if from_email is None:
        from_email = saas_settings.DEFAULT_FROM_EMAIL
    text_message, html_message = render_mail_messages(template_id, context)
    result: TaskResult = send_email.enqueue(
        subject=str(subject),
        recipients=recipients,
        text_message=text_message,
        html_message=html_message,
        from_email=from_email,
        headers=headers,
        reply_to=reply_to,
    )
    mail_queued.send(
        sender=None,
        template_id=template_id,
        **result.kwargs,
        result=result,
    )


@task
def send_email(
    subject: str,
    recipients: t.List[str],
    text_message: str,
    html_message: t.Optional[str] = None,
    from_email: t.Optional[str] = None,
    headers: t.Optional[t.Dict[str, str]] = None,
    reply_to: t.Optional[str] = None,
):
    mail = EmailMultiAlternatives(
        subject,
        body=text_message,
        from_email=from_email,
        to=recipients,
        headers=headers,
        reply_to=reply_to,
    )
    if html_message:
        mail.attach_alternative(html_message, 'text/html')

    return mail.send()


def render_mail_messages(template_id: str, context: t.Dict[str, t.Any]) -> t.Tuple[str, str]:
    context.setdefault('site', saas_settings.SITE)
    text: str = loader.render_to_string(f'saas_emails/{template_id}.text', context)
    html: str = loader.render_to_string(f'saas_emails/{template_id}.html', context)
    return text, css_inline.inline(html)



from .db.migrate_receiver import create_permissions_receiver


DEFAULT_PERMISSIONS = [
    ('tenant.read', 'Read permission for tenants'),
    ('tenant.write', 'Write permission for tenants'),
    ('tenant.admin', 'Admin permission for tenants'),
]

create_permissions = create_permissions_receiver(DEFAULT_PERMISSIONS)



from email.utils import formataddr
from django.utils.translation import gettext_lazy as _
from django.conf import settings
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.throttling import AnonRateThrottle
from saas_base.drf.views import Endpoint
from saas_base.settings import saas_settings
from saas_base.models import UserEmail
from saas_base.rules import check_rules
from saas_base.serializers.password import (
    PasswordForgetSerializer,
    PasswordResetSerializer,
)
from saas_base.tasks.send_mails import send_template_email

__all__ = [
    'PasswordForgotEndpoint',
    'PasswordResetEndpoint',
]


class PasswordForgotEndpoint(Endpoint):
    email_template_id = 'reset_password'
    email_subject = _('Password Reset Request')

    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordForgetSerializer

    def post(self, request: Request):
        """Send a forgot password reset email code."""
        serializer: PasswordForgetSerializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj: UserEmail = serializer.save()

        # check bad request rules
        check_rules(saas_settings.RESET_PASSWORD_SECURITY_RULES, request)

        code = serializer.save_auth_code(obj.user_id)
        name = obj.user.get_full_name() or obj.user.get_username()
        recipients = [formataddr((name, obj.email))]
        send_template_email(
            template_id=self.email_template_id,
            subject=self.email_subject,
            recipients=recipients,
            context={'code': code, 'user': obj.user},
        )
        return Response(status=204)


class PasswordResetEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordResetSerializer

    def post(self, request: Request):
        """Reset password of a user with the given code."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({'next': settings.LOGIN_URL})



TESTING = True
SECRET_KEY = 'django-insecure'
ALLOWED_HOSTS = ['*']
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}
TASKS = {
    'default': {
        'BACKEND': 'django.tasks.backends.immediate.ImmediateBackend',
    },
}
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
            ]
        },
    }
]
AUTHENTICATION_BACKENDS = [
    'saas_base.auth.backends.ModelBackend',
]
MIDDLEWARE = [
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'saas_base.middleware.HeaderTenantIdMiddleware',
    'saas_base.middleware.PathTenantIdMiddleware',
    'saas_base.middleware.SessionTenantIdMiddleware',
    'saas_base.middleware.TenantMiddleware',
]
SESSION_ENGINE = 'django.contrib.sessions.backends.cache'
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'unique-snowflake',
    }
}
EMAIL_BACKEND = 'django.core.mail.backends.locmem.EmailBackend'
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
        'OPTIONS': {
            'min_length': 8,
        },
    },
]
INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.sessions',
    'rest_framework',
    'drf_spectacular',
    'saas_base',
    'saas_base.drf',
]
REST_FRAMEWORK = {
    'PAGE_SIZE': 10,
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    'DEFAULT_RENDERER_CLASSES': ['rest_framework.renderers.JSONRenderer'],
    'DEFAULT_PARSER_CLASSES': ['rest_framework.parsers.JSONParser'],
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.PageNumberPagination',
    'DEFAULT_SCHEMA_CLASS': 'saas_base.drf.spectacular.AutoSchema',
}
USE_TZ = True
TIME_ZONE = 'UTC'
ROOT_URLCONF = 'tests.urls'



from rest_framework.mixins import (
    RetrieveModelMixin,
    CreateModelMixin,
    UpdateModelMixin,
    ListModelMixin,
    DestroyModelMixin,
)
from rest_framework.request import Request
from ..drf.views import TenantEndpoint
from ..models import Group
from ..serializers.group import GroupSerializer


__all__ = ['GroupListEndpoint', 'GroupItemEndpoint']


class GroupListEndpoint(ListModelMixin, CreateModelMixin, TenantEndpoint):
    serializer_class = GroupSerializer
    queryset = Group.objects.all()
    pagination_class = None
    resource_scopes = ['tenant', 'tenant:group']

    def get(self, request: Request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request: Request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def perform_create(self, serializer):
        tenant_id = self.get_tenant_id()
        serializer.save(tenant_id=tenant_id)


class GroupItemEndpoint(RetrieveModelMixin, UpdateModelMixin, DestroyModelMixin, TenantEndpoint):
    serializer_class = GroupSerializer
    queryset = Group.objects.all()
    resource_scopes = ['tenant', 'tenant:group']

    def get(self, request: Request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request: Request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request: Request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)

    def delete(self, request: Request, *args, **kwargs):
        """Remove a permission group."""
        return self.destroy(request, *args, **kwargs)



from rest_framework.request import Request
from rest_framework.mixins import (
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
)
from ..drf.filters import IncludeFilter, ChoiceFilter
from ..drf.views import AuthenticatedEndpoint
from ..models import Member
from ..serializers.member import UserMembershipSerializer

__all__ = [
    'UserMemberListEndpoint',
    'UserMemberItemEndpoint',
]


class UserMemberListEndpoint(ListModelMixin, AuthenticatedEndpoint):
    serializer_class = UserMembershipSerializer
    queryset = Member.objects.select_related('tenant').all()
    filter_backends = [IncludeFilter, ChoiceFilter]
    resource_scopes = ['user', 'user:member']

    choice_filter_fields = ['status']
    include_select_related_fields = ['role']
    include_prefetch_related_fields = ['groups', 'permissions', 'groups__permissions']

    def get_queryset(self):
        return self.queryset.filter(user=self.request.user)

    def get(self, request: Request, *args, **kwargs):
        """List all the current user's tenants."""
        return self.list(request, *args, **kwargs)


class UserMemberItemEndpoint(RetrieveModelMixin, UpdateModelMixin, DestroyModelMixin, AuthenticatedEndpoint):
    serializer_class = UserMembershipSerializer
    queryset = Member.objects.all()
    resource_scopes = ['user', 'user:member']

    def get_queryset(self):
        return self.queryset.filter(user=self.request.user)

    def get(self, request: Request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def patch(self, request: Request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)

    def delete(self, request: Request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)



from django.urls import path
from ..endpoints.members import (
    MemberListEndpoint,
    MemberItemEndpoint,
)

urlpatterns = [
    path('', MemberListEndpoint.as_view()),
    path('<pk>/', MemberItemEndpoint.as_view()),
]



from django.conf import settings
from rest_framework.fields import BooleanField
from rest_framework.filters import BaseFilterBackend
from rest_framework.exceptions import ValidationError
from .errors import BadRequest
from ..settings import saas_settings

__all__ = [
    'TenantIdFilter',
    'IncludeFilter',
    'ChoiceFilter',
]


class TenantIdFilter(BaseFilterBackend):
    tenant_id_field = 'tenant_id'

    def filter_queryset(self, request, queryset, view):
        query_field = getattr(view, 'tenant_id_field', self.tenant_id_field)

        tenant_id = getattr(request, 'tenant_id', None)
        if not tenant_id:
            raise BadRequest('Missing Tenant ID')

        kwargs = {query_field: tenant_id}
        return queryset.filter(**kwargs)

    def get_schema_operation_parameters(self, view):
        in_header = 'saas_base.middleware.HeaderTenantIdMiddleware' in settings.MIDDLEWARE
        in_path = 'saas_base.middleware.PathTenantIdMiddleware' in settings.MIDDLEWARE

        parameters = []
        if in_header:
            parameters.append(
                {
                    'name': saas_settings.TENANT_ID_HEADER,
                    'required': True,
                    'in': 'header',
                    'schema': {
                        'type': 'string',
                    },
                }
            )

        if in_path:
            parameters.append(
                {
                    'name': 'tenant_id',
                    'required': True,
                    'in': 'path',
                    'schema': {
                        'type': 'string',
                    },
                }
            )

        if len(parameters) == 2:
            parameters[0]['required'] = False
            parameters[0]['description'] = 'Tenant ID (required if not provided in path).'
            parameters[1]['required'] = False
            parameters[1]['description'] = 'Tenant ID (required if not provided in header).'

        return parameters


class IncludeFilter(BaseFilterBackend):
    @staticmethod
    def get_select_related_fields(view):
        return getattr(view, 'include_select_related_fields', [])

    @staticmethod
    def get_prefetch_related_fields(view):
        return getattr(view, 'include_prefetch_related_fields', [])

    @staticmethod
    def get_annotate_fields(view):
        return getattr(view, 'include_annotate_fields', [])

    @staticmethod
    def get_include_terms(request):
        params = request.query_params.get('include', '')
        params = params.replace('\x00', '')  # strip null characters
        params = params.replace(',', ' ')
        return params.split()

    @staticmethod
    def annotate_queryset(queryset, terms, view):
        func = getattr(view, 'annotate_queryset', None)
        if func:
            return func(queryset, terms)
        return queryset

    def filter_queryset(self, request, queryset, view):
        select_related_fields = self.get_select_related_fields(view)
        prefetch_related_fields = self.get_prefetch_related_fields(view)
        annotate_fields = self.get_annotate_fields(view)
        if not select_related_fields and not prefetch_related_fields and not annotate_fields:
            return queryset

        include_terms = self.get_include_terms(request)
        if not include_terms:
            return queryset

        if include_terms == ['all']:
            include_terms = select_related_fields + prefetch_related_fields + annotate_fields

        annotate_terms = []
        for field in include_terms:
            if field in select_related_fields:
                queryset = queryset.select_related(field)
            elif field in prefetch_related_fields:
                queryset = queryset.prefetch_related(field)
                relations = [key for key in prefetch_related_fields if key.startswith(f'{field}__')]
                queryset = queryset.prefetch_related(*relations)
            elif field in annotate_fields:
                annotate_terms.append(field)

        if annotate_terms:
            queryset = self.annotate_queryset(queryset, annotate_terms, view)

        request.include_terms = include_terms
        return queryset

    def get_schema_operation_parameters(self, view):
        select_related_fields = self.get_select_related_fields(view)
        prefetch_related_fields = self.get_prefetch_related_fields(view)
        annotate_fields = self.get_annotate_fields(view)
        related_fields = ['all'] + select_related_fields + prefetch_related_fields + annotate_fields
        desc = ', '.join([f'`"{name}"`' for name in related_fields if '__' not in name])
        return [
            {
                'name': 'include',
                'required': False,
                'in': 'query',
                'description': f'Include related fields of {desc}',
                'schema': {
                    'type': 'string',
                },
            },
        ]


class ChoiceFilter(BaseFilterBackend):
    default_choice_filter_fields = ['status']

    def get_choice_filter_fields(self, view):
        return getattr(view, 'choice_filter_fields', self.default_choice_filter_fields)

    def get_choice_filter_terms(self, view, request):
        params = {}
        for key in self.get_choice_filter_fields(view):
            value = request.query_params.get(key, '')
            if value:
                params[key] = value
        return params

    def filter_queryset(self, request, queryset, view):
        terms = self.get_choice_filter_terms(view, request)
        if not terms:
            return queryset

        serializer_cls = view.get_serializer_class()
        serializer = serializer_cls()
        for key in terms:
            field = serializer.fields[key]
            try:
                value = field.to_internal_value(terms[key])
            except ValidationError:
                continue
            queryset = queryset.filter(**{key: value})
        return queryset

    def get_schema_operation_parameters(self, view):
        choice_filter_fields = self.get_choice_filter_fields(view)
        serializer_cls = view.get_serializer_class()
        serializer = serializer_cls()
        schema = []

        for key in choice_filter_fields:
            field = serializer.fields[key]
            if isinstance(field, BooleanField):
                schema.append(
                    {
                        'name': key,
                        'required': False,
                        'in': 'query',
                        'schema': {
                            'type': 'boolean',
                        },
                    }
                )
            else:
                choices = [field.to_representation(v) for v in field.choices]
                schema.append(
                    {
                        'name': key,
                        'required': False,
                        'in': 'query',
                        'schema': {
                            'type': 'string',
                            'enum': choices,
                        },
                    }
                )
        return schema



from rest_framework import status
from rest_framework.exceptions import APIException

__all__ = ['BadRequest']


class BadRequest(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'Bad request.'
    default_code = 'bad_request'



from django.http import HttpResponse
from django.test import RequestFactory
from saas_base.middleware import TenantMiddleware
from saas_base.test import SaasTestCase


class TestTenantMiddleware(SaasTestCase):
    middleware_cls = TenantMiddleware

    @staticmethod
    def create_request(host):
        meta = {'HTTP_HOST': host}
        factory = RequestFactory(**meta)
        return factory.get('/')

    def setUp(self):
        self.middleware = self.middleware_cls(lambda req: HttpResponse())

    def test_match_cached_tenant(self):
        request = self.create_request('none.us.localhost')
        request.tenant_id = 1
        self.middleware(request)
        self.assertEqual(request.tenant, self.tenant)

    def test_missing_tenant_id(self):
        request = self.create_request('none.us.localhost')
        self.middleware(request)
        self.assertEqual(request.tenant, None)

    def test_not_found_tenant(self):
        request = self.create_request('none.us.localhost')
        request.tenant_id = 404
        self.middleware(request)
        self.assertEqual(request.tenant, None)



from django.urls import path, include
from drf_spectacular.views import SpectacularJSONAPIView


urlpatterns = [
    path('m/', include('saas_base.api_urls.all')),
    path('tenant/<tenant_id>/members/', include('saas_base.api_urls.members')),
    path('schema/openapi', SpectacularJSONAPIView.as_view(), name='schema'),
]



import pathlib
from functools import cache
from rest_framework.request import Request
from .base import Rule


@cache
def load_default_blocked_domains():
    with pathlib.Path(__file__).parent.joinpath('blocklist.txt').open() as f:
        return f.read().splitlines()


class BlockedEmailDomains(Rule):
    """This rule prevents using email addresses from blocked domains.
    For example, `username@boofx.com` is not allowed.
    """
    def bad_request(self, request: Request):
        blocked_list = self.options.get('domains')
        if not blocked_list:
            blocked_list = load_default_blocked_domains()

        email = self.get_response_field_value(request)
        return blocked_list and email.endswith(tuple([f'@{s}' for s in blocked_list]))


class AvoidTooManyDots(Rule):
    """This rule prevents using too many dots in email address.
    For example, `a.b.c.d.e@gmail.com` is not allowed.
    """
    MAX_DOT_COUNT = 4

    def bad_request(self, request: Request):
        max_dot_count = self.options.get('count', self.MAX_DOT_COUNT)
        email = self.get_response_field_value(request)
        name = email.split('@')[0]
        return name.count('.') > max_dot_count


class AvoidUsingPlus(Rule):
    """This rule prevents using of `+` in email address.
    For example, `username+random@gmail.com` is not allowed.
    """
    def bad_request(self, request: Request):
        email = self.get_response_field_value(request)
        return '+' in email



import warnings
from saas_base.rules import check_rules as check_security_rules
from .rules import Rule, BlockedEmailDomains, TooManyDots, Turnstile

warnings.warn('saas_base.security is deprecated, use saas_base.rules instead', DeprecationWarning)

__all__ = [
    'Rule',
    'BlockedEmailDomains',
    'TooManyDots',
    'Turnstile',
    'check_security_rules',
]
